import jax

jax.config.update("jax_enable_x64", True)

from jaxlayerlumos.jaxlayerlumos import stackrt_n_k
from jaxlayerlumos.jaxlayerlumos import stackrt_eps_mu

from jaxlayerlumos.wrappers import stackrt_n_k_0
from jaxlayerlumos.wrappers import stackrt_n_k_45


stackrt = stackrt_n_k
stackrt0 = stackrt_n_k_0
stackrt45 = stackrt_n_k_45



import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import stackrt_eps_mu
from jaxlayerlumos import utils_materials


def test_stackrt_radar_1():
    frequencies = jnp.linspace(0.1e9, 1e9, 3)
    # frequencies = jnp.array([1e9])
    materials = onp.array(["Air", "11", "16", "7", "4", "4", "PEC"])
    eps_stack, mu_stack = utils_materials.get_eps_mu(materials, frequencies)

    d_stack = jnp.array([0, 0.7742, 0.8485, 1.4878, 1.9883, 1.9863, 0]) * 1e-3
    R_TE, T_TE, R_TM, T_TM = stackrt_eps_mu(
        eps_stack, mu_stack, d_stack, frequencies, 0.0
    )

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    R_db = 10 * jnp.log10(R_avg).squeeze()
    T_db = 10 * jnp.log10(T_avg).squeeze()

    expected_R_db = onp.array(
        [
            -33.872113339369314,
            -39.141066380663602,
            -38.794125046637177,
        ]
    )

    expected_T_db = onp.array(
        [
            -onp.inf,
            -onp.inf,
            -onp.inf,
        ]
    )

    print("R_db")
    for elem in R_db:
        print(elem)

    print("T_db")
    for elem in T_db:
        print(elem)

    onp.testing.assert_allclose(R_db, expected_R_db)
    onp.testing.assert_allclose(T_db, expected_T_db)


def test_stackrt_radar_2():
    frequencies = jnp.linspace(0.1e9, 1e9, 5)

    materials = onp.array(["Air", "15", "12", "9", "PEC"])
    eps_stack, mu_stack = utils_materials.get_eps_mu(materials, frequencies)

    d_stack = jnp.array([0, 0.43, 0.21, 0.05, 0]) * 1e-3
    R_TE, T_TE, R_TM, T_TM = stackrt_eps_mu(
        eps_stack, mu_stack, d_stack, frequencies, 0.0
    )

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    R_db = 10 * jnp.log10(R_avg).squeeze()
    T_db = 10 * jnp.log10(T_avg).squeeze()

    expected_R_db = onp.array(
        [
            -0.057683530802195224,
            -0.5144977473090127,
            -1.1825636270341442,
            -1.9171267047938356,
            -2.673613156694648,
        ]
    )

    expected_T_db = onp.array(
        [
            -onp.inf,
            -onp.inf,
            -onp.inf,
            -onp.inf,
            -onp.inf,
        ]
    )

    print("R_db")
    for elem in R_db:
        print(elem)

    print("T_db")
    for elem in T_db:
        print(elem)

    onp.testing.assert_allclose(R_db, expected_R_db)
    onp.testing.assert_allclose(T_db, expected_T_db)



import jax
import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import utils_materials
from jaxlayerlumos import utils_units


def true_fun(r_jk_TE, t_jk_TE, r_jk_TM, t_jk_TM):
    r_jk_TE = r_jk_TE.at[-1].set(-1.0)
    t_jk_TE = t_jk_TE.at[-1].set(1.0)
    r_jk_TM = r_jk_TM.at[-1].set(-1.0)
    t_jk_TM = t_jk_TM.at[-1].set(1.0)

    return r_jk_TE, t_jk_TE, r_jk_TM, t_jk_TM


def false_fun(r_jk_TE, t_jk_TE, r_jk_TM, t_jk_TM):
    return r_jk_TE, t_jk_TE, r_jk_TM, t_jk_TM


def stackrt_eps_mu_base(eps_r, mu_r, thicknesses, f_i, thetas_k, return_coeffs=False):
    assert isinstance(eps_r, jnp.ndarray)
    assert isinstance(mu_r, jnp.ndarray)
    assert isinstance(thicknesses, jnp.ndarray)

    assert eps_r.ndim == 1
    assert mu_r.ndim == 1
    assert thicknesses.ndim == 1

    assert thicknesses[0] == 0
    assert thicknesses[-1] == 0

    assert eps_r.shape[0] == thicknesses.shape[0]
    assert mu_r.shape[0] == thicknesses.shape[0]

    c = utils_units.get_light_speed()
    n = jnp.conj(jnp.sqrt(eps_r * mu_r))
    k = 2 * jnp.pi / c * f_i * n
    eta = jnp.conj(jnp.sqrt(mu_r / eps_r))

    sin_theta = jnp.expand_dims(jnp.sin(thetas_k), axis=0)
    sin_theta = sin_theta * n[0] / n
    cos_theta_t = jnp.sqrt(1 - sin_theta**2)
    kz = k * cos_theta_t

    upper_bound = 600.0
    delta = thicknesses * kz
    delta = jnp.real(delta) + 1j * jnp.clip(jnp.imag(delta), -upper_bound, upper_bound)

    Z_TE = eta / cos_theta_t
    Z_TM = eta * cos_theta_t

    r_jk_TE = (Z_TE[1:] - Z_TE[:-1]) / (Z_TE[1:] + Z_TE[:-1])
    t_jk_TE = (2 * Z_TE[1:]) / (Z_TE[1:] + Z_TE[:-1])

    r_jk_TM = (Z_TM[1:] - Z_TM[:-1]) / (Z_TM[1:] + Z_TM[:-1])
    t_jk_TM = (
        (2 * Z_TM[1:]) / (Z_TM[1:] + Z_TM[:-1]) * cos_theta_t[:-1] / cos_theta_t[1:]
    )

    r_jk_TE, t_jk_TE, r_jk_TM, t_jk_TM = jax.lax.cond(
        jnp.isinf(jnp.real(eps_r[-1]))
        & jnp.isclose(jnp.imag(eps_r[-1]), 0)
        & jnp.isclose(jnp.real(mu_r[-1]), 1)
        & jnp.isclose(jnp.imag(mu_r[-1]), 0),
        true_fun,
        false_fun,
        r_jk_TE,
        t_jk_TE,
        r_jk_TM,
        t_jk_TM,
    )

    t_inv_TE = 1 / t_jk_TE
    r_over_t_TE = r_jk_TE / t_jk_TE

    D_TE = jnp.stack(
        [
            jnp.stack([t_inv_TE, r_over_t_TE], axis=-1),
            jnp.stack([r_over_t_TE, t_inv_TE], axis=-1),
        ],
        axis=-2,
    )

    t_inv_TM = 1 / t_jk_TM
    r_over_t_TM = r_jk_TM / t_jk_TM

    D_TM = jnp.stack(
        [
            jnp.stack([t_inv_TM, r_over_t_TM], axis=-1),
            jnp.stack([r_over_t_TM, t_inv_TM], axis=-1),
        ],
        axis=-2,
    )

    exp_neg_jdelta = jnp.exp(-1j * delta[0:-1])
    exp_pos_jdelta = jnp.exp(1j * delta[0:-1])

    zeros = jnp.zeros_like(exp_neg_jdelta)

    P = jnp.stack(
        [
            jnp.stack([exp_neg_jdelta, zeros], axis=-1),
            jnp.stack([zeros, exp_pos_jdelta], axis=-1),
        ],
        axis=-2,
    )

    DP_TE = jnp.matmul(P, D_TE)
    DP_TM = jnp.matmul(P, D_TM)

    if not return_coeffs:

        def matmul_scan(a, b):
            return jnp.matmul(a, b)

        M_TE = jax.lax.associative_scan(matmul_scan, DP_TE)[-1]
        M_TM = jax.lax.associative_scan(matmul_scan, DP_TM)[-1]
    else:

        def matmul_left(a, b):
            return jnp.matmul(b, a)

        DP_TE_rev = jnp.flip(DP_TE, axis=0)

        M_TE_all = jax.lax.associative_scan(matmul_left, DP_TE_rev)
        M_TE_flipped = jnp.flip(M_TE_all, axis=0)

        M_TE = M_TE_all[-1]

        coeff_TE = M_TE_flipped[:, :, 0]
        coeff_TE = jnp.concatenate((coeff_TE, jnp.array([[1, 0]])), axis=0)
        coeff_TE = coeff_TE / M_TE[0, 0]

        DP_TM_rev = jnp.flip(DP_TM, axis=0)

        M_TM_all = jax.lax.associative_scan(matmul_left, DP_TM_rev)
        M_TM_flipped = jnp.flip(M_TM_all, axis=0)
        M_TM = M_TM_all[-1]

        coeff_TM = M_TM_flipped[:, :, 0]
        coeff_TM = jnp.concatenate((coeff_TM, jnp.array([[1, 0]])), axis=0)
        coeff_TM = coeff_TM / M_TM[0, 0]

        results_coeffs = {
            "coeff_TE": coeff_TE,
            "coeff_TM": coeff_TM,
            "n": n,
            "k_z": kz,
            "cos_theta": cos_theta_t,
        }

    r_TE_i = M_TE[1, 0] / M_TE[0, 0]
    t_TE_i = 1 / M_TE[0, 0]

    r_TM_i = M_TM[1, 0] / M_TM[0, 0]
    t_TM_i = 1 / M_TM[0, 0]

    R_TE = jnp.abs(r_TE_i) ** 2
    R_TM = jnp.abs(r_TM_i) ** 2

    T_TE = (
        jnp.abs(t_TE_i) ** 2
        * jnp.real(n[-1] * cos_theta_t[-1])
        / jnp.real(n[0] * cos_theta_t[0])
    )
    T_TM = (
        jnp.abs(t_TM_i) ** 2
        * jnp.real(n[-1] * jnp.conj(cos_theta_t[-1]))
        / jnp.real(n[0] * jnp.conj(cos_theta_t[0]))
    )

    if return_coeffs:
        power_entering_TE = jnp.real(
            n[0] * cos_theta_t[0] * (1 + jnp.conj(r_TE_i)) * (1 - r_TE_i)
        ) / jnp.real(n[0] * cos_theta_t[0])

        power_entering_TM = jnp.real(
            n[0] * jnp.conj(cos_theta_t[0]) * (1 + r_TM_i) * (1 - jnp.conj(r_TM_i))
        ) / jnp.real(n[0] * jnp.conj(cos_theta_t[0]))

        results_coeffs.update(
            {
                "power_entering_TE": power_entering_TE,
                "power_entering_TM": power_entering_TM,
                "R_TE": R_TE,
                "T_TE": T_TE,
                "R_TM": R_TM,
                "T_TM": T_TM,
            }
        )
        return R_TE, T_TE, R_TM, T_TM, results_coeffs
    else:
        return R_TE, T_TE, R_TM, T_TM


def stackrt_eps_mu_theta(eps_r, mu_r, d, f, theta, return_coeffs=False):
    assert isinstance(eps_r, jnp.ndarray)
    assert isinstance(mu_r, jnp.ndarray)
    assert isinstance(d, jnp.ndarray)
    assert isinstance(f, jnp.ndarray)

    assert eps_r.ndim == 2
    assert mu_r.ndim == 2
    assert d.ndim == 1
    assert f.ndim == 1

    assert eps_r.shape[0] == f.shape[0]
    assert eps_r.shape[1] == d.shape[0]

    assert mu_r.shape[0] == f.shape[0]
    assert mu_r.shape[1] == d.shape[0]

    theta_rad = jnp.radians(theta)

    if not return_coeffs:
        fun_mapped = jax.vmap(stackrt_eps_mu_base, (0, 0, None, 0, None), (0, 0, 0, 0))

        R_TE, T_TE, R_TM, T_TM = fun_mapped(eps_r, mu_r, d, f, theta_rad)
    else:
        fun_mapped = jax.vmap(
            stackrt_eps_mu_base, (0, 0, None, 0, None, None), (0, 0, 0, 0, 0)
        )

        R_TE, T_TE, R_TM, T_TM, results_coeffs = fun_mapped(
            eps_r, mu_r, d, f, theta_rad, return_coeffs
        )

    if (
        jnp.all(jnp.isinf(jnp.real(eps_r[:, -1])))
        and jnp.allclose(jnp.imag(eps_r[:, -1]), 0)
        and jnp.allclose(jnp.real(mu_r[:, -1]), 1)
        and jnp.allclose(jnp.imag(mu_r[:, -1]), 0)
    ):  # TODO: it is needed? It is just enforcing outputs. Outputs should be zeros by calculation.
        T_TE = jnp.zeros_like(R_TE)
        T_TM = jnp.zeros_like(R_TM)

    if not return_coeffs:
        return R_TE, T_TE, R_TM, T_TM
    else:
        return R_TE, T_TE, R_TM, T_TM, results_coeffs


def stackrt_eps_mu(eps_r, mu_r, d, f, thetas):
    if thetas is None:
        thetas = jnp.array([0])
    elif isinstance(thetas, (float, int)):
        thetas = jnp.array([thetas])

    fun_mapped = jax.vmap(
        stackrt_eps_mu_theta, (None, None, None, None, 0), (0, 0, 0, 0)
    )
    R_TE, T_TE, R_TM, T_TM = fun_mapped(eps_r, mu_r, d, f, thetas)

    return R_TE, T_TE, R_TM, T_TM


def stackrt_n_k(
    refractive_indices, thicknesses, frequencies, thetas, return_coeffs=False
):
    """
    Calculate reflection and transmission coefficients for a multilayer stack at
    different frequencies and incidence angles.

    Parameters:
    - refractive_indices: The refractive indices of the layers for each frequency.
        Shape should be (Nfreq, Nlayers), where Nfreq is the number of frequencies
        and Nlayers is the number of layers.
    - thicknesses: The thicknesses of the layers. Shape should be (Nlayers, ).
    - frequencies: The frequencies at which to calculate the coefficients. Shape
        should be (Nfreq, ).
    - thetas: The incidence angle(s) in degrees. Can be a single value or an array
        of angles. If it is None, it will be [0].

    Returns:
    - A tuple containing:
      - R_TE (jax.numpy.ndarray): Reflectance for TE polarization.
            Shape is (Nfreq, ).
      - T_TE (jax.numpy.ndarray): Transmittance for TE polarization.
            Shape is (Nfreq, ).
      - R_TM (jax.numpy.ndarray): Reflectance for TM polarization.
            Shape is (Nfreq, ).
      - T_TM (jax.numpy.ndarray): Transmittance for TM polarization.
            Shape is (Nfreq, ).

    """

    assert isinstance(refractive_indices, jnp.ndarray)
    assert isinstance(thicknesses, jnp.ndarray)
    assert isinstance(frequencies, jnp.ndarray)
    assert isinstance(thetas, (type(None), jnp.ndarray, float, int))

    assert refractive_indices.ndim == 2
    assert thicknesses.ndim == 1
    assert frequencies.ndim == 1

    assert refractive_indices.shape[0] == frequencies.shape[0]
    assert refractive_indices.shape[1] == thicknesses.shape[0]

    if thetas is None:
        thetas = jnp.array([0])
    elif isinstance(thetas, (float, int)):
        thetas = jnp.array([thetas])

    eps_r, mu_r = utils_materials.convert_n_k_to_eps_mu_for_non_magnetic_materials(
        refractive_indices
    )

    if not return_coeffs:
        fun_mapped = jax.vmap(
            stackrt_eps_mu_theta, (None, None, None, None, 0), (0, 0, 0, 0)
        )
        R_TE, T_TE, R_TM, T_TM = fun_mapped(
            eps_r, mu_r, thicknesses, frequencies, thetas
        )
        return R_TE, T_TE, R_TM, T_TM
    else:
        fun_mapped = jax.vmap(
            stackrt_eps_mu_theta, (None, None, None, None, 0, None), (0, 0, 0, 0, 0)
        )
        R_TE, T_TE, R_TM, T_TM, results_coeffs = fun_mapped(
            eps_r, mu_r, thicknesses, frequencies, thetas, return_coeffs
        )
        return R_TE, T_TE, R_TM, T_TM, results_coeffs



import jax.numpy as jnp


Metals_sigma = {
    "Cu": 5.96e7,
    "Cr": 7.74e6,
    "Ag": 6.3e7,
    "Al": 3.77e7,
    "Ni": 1.43e7,
    "W": 1.79e7,
    "Ti": 2.38e6,
    "Pd": 9.52e6,
}

Metals_nk_updated_specific_sigma = {}
mu0 = 4 * jnp.pi * 1e-7  # H/m
Z_0 = 377  # Ohms, impedance of free space
# Frequency range
nu = jnp.linspace(8e9, 18e9, 11)  # From 8 GHz to 18 GHz
omega = 2 * jnp.pi * nu  # Angular frequency

for metal, sigma in Metals_sigma.items():
    Z = jnp.sqrt(
        1j * omega * mu0 / sigma
    )  # Impedance of the material using specific sigma
    n_complex = Z_0 / Z  # Complex refractive index

    # Extract real and imaginary parts of the refractive index
    n_real = jnp.real(n_complex)
    k_imag = jnp.imag(n_complex)

    # Update the dictionary with the new values
    Metals_nk_updated_specific_sigma[metal] = {
        "freq_data": nu.tolist(),  # Converting JAX arrays to lists for JSON serialization
        "n_data": n_real.tolist(),
        "k_data": k_imag.tolist(),
    }


def load_material_RF(material_name, frequencies):
    """
    Load material RF data for a given material and frequencies. Adapted for JAX.

    Parameters:
    - material_name: The name of the material to load.
    - frequencies: Array of frequencies for which data is requested.

    Returns:
    - A JAX array with columns for frequency, n, and k.
    """

    frequencies = jnp.array(frequencies)  # Ensure input frequencies are JAX arrays

    if material_name not in Metals_nk_updated_specific_sigma:
        n_default = jnp.ones_like(frequencies)
        k_default = jnp.zeros_like(frequencies)
        data = jnp.column_stack((frequencies, n_default, k_default))
    else:
        material_data = Metals_nk_updated_specific_sigma[material_name]
        freq_data = jnp.array(material_data["freq_data"])
        n_data = jnp.array(material_data["n_data"])
        k_data = jnp.array(material_data["k_data"])

        n_interpolated = jnp.interp(
            frequencies, freq_data, n_data, left="extrapolate", right="extrapolate"
        )
        k_interpolated = jnp.interp(
            frequencies, freq_data, k_data, left="extrapolate", right="extrapolate"
        )

        data = jnp.column_stack((frequencies, n_interpolated, k_interpolated))

    return data



import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import stackrt
from jaxlayerlumos import utils_materials
from jaxlayerlumos import utils_spectra
from jaxlayerlumos import utils_layers
from jaxlayerlumos import utils_units


def test_sizes():
    num_wavelengths = 123
    num_angles = 4

    frequencies = utils_spectra.get_frequencies_wide_visible_light(
        num_wavelengths=num_wavelengths
    )
    materials = ["TiO2"]
    n_stack = utils_materials.get_n_k_surrounded_by_air(materials, frequencies)
    d_stack = utils_layers.get_thicknesses_surrounded_by_air(jnp.array([2e-8]))
    thetas = jnp.linspace(0, 89, num_angles)

    R_TE, T_TE, R_TM, T_TM = stackrt(n_stack, d_stack, frequencies, thetas)

    assert isinstance(R_TE, jnp.ndarray)
    assert isinstance(R_TM, jnp.ndarray)
    assert isinstance(T_TE, jnp.ndarray)
    assert isinstance(T_TM, jnp.ndarray)
    assert R_TE.ndim == 2
    assert R_TM.ndim == 2
    assert T_TE.ndim == 2
    assert T_TM.ndim == 2
    assert R_TE.shape[0] == R_TM.shape[0] == num_angles
    assert T_TE.shape[0] == T_TM.shape[0] == num_angles
    assert R_TE.shape[1] == R_TM.shape[1] == num_wavelengths
    assert T_TE.shape[1] == T_TM.shape[1] == num_wavelengths


def test_angles_1():
    num_wavelengths = 5
    frequencies = utils_spectra.get_frequencies_visible_light(
        num_wavelengths=num_wavelengths
    )

    materials = ["Air", "TiO2", "Air"]
    d_stack = jnp.array([0, 2e-8, 0])
    thetas = jnp.linspace(0, 89, 3)
    n_k = utils_materials.get_n_k(materials, frequencies)

    R_TE, T_TE, R_TM, T_TM = stackrt(n_k, d_stack, frequencies, thetas)

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    expected_R_avg = jnp.array(
        [
            [
                0.4770790218584379,
                0.2918301244526284,
                0.20026632179451573,
                0.14616448215891184,
                0.11132762140037065,
            ],
            [
                0.4558617206305765,
                0.2910941832925791,
                0.20684237462156152,
                0.15476340159980778,
                0.11998132563573674,
            ],
            [
                0.9903190571800133,
                0.9857622632656011,
                0.979727833282493,
                0.9726668981700657,
                0.9646925772238752,
            ],
        ]
    )

    expected_T_avg = jnp.array(
        [
            [
                0.5198973861437665,
                0.7081681748619181,
                0.7997336666045016,
                0.8538355175673582,
                0.8886723785996291,
            ],
            [
                0.5410727876996986,
                0.7089041021153872,
                0.793157613579433,
                0.8452365981191836,
                0.8800186743642635,
            ],
            [
                0.009481699666494146,
                0.014237605935075181,
                0.020272165555583664,
                0.027333101794498202,
                0.035307422776128854,
            ],
        ]
    )

    print("R_avg")
    for elem_1 in R_avg:
        for elem_2 in elem_1:
            print(elem_2)

    print("T_avg")
    for elem_1 in T_avg:
        for elem_2 in elem_1:
            print(elem_2)

    onp.testing.assert_allclose(R_avg, expected_R_avg)
    onp.testing.assert_allclose(T_avg, expected_T_avg)


def test_angles_2():
    wavelengths = jnp.array([300e-9])
    frequencies = utils_units.get_light_speed() / wavelengths

    materials = onp.array(["Air", "FusedSilica", "Si3N4"])
    thickness_materials = [0, 2.91937911, 0]
    theta = 47.1756

    thicknesses = jnp.array(thickness_materials)
    n_k = utils_materials.get_n_k(materials, frequencies)

    thicknesses *= utils_units.get_nano()

    R_TE, T_TE, R_TM, T_TM = stackrt(n_k, thicknesses, frequencies, theta)

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    print("R_TE")
    for elem_1 in R_TE:
        for elem_2 in elem_1:
            print(elem_2)

    print("T_TE")
    for elem_1 in T_TE:
        for elem_2 in elem_1:
            print(elem_2)

    print("R_TM")
    for elem_1 in R_TM:
        for elem_2 in elem_1:
            print(elem_2)

    print("T_TM")
    for elem_1 in T_TM:
        for elem_2 in elem_1:
            print(elem_2)

    expected_R_avg = jnp.array([[0.24877806962502294 + 0.048295322372087564]]) / 2
    expected_T_avg = jnp.array([[0.7512219303749771 + 0.9517046776279122]]) / 2

    onp.testing.assert_allclose(R_avg, expected_R_avg)
    onp.testing.assert_allclose(T_avg, expected_T_avg)


def test_angles_3():
    wavelengths = jnp.array([300e-9])
    frequencies = utils_units.get_light_speed() / wavelengths

    materials = onp.array(["Air", "Ag", "Cr"])
    thickness_materials = [0, 9.36793259, 0.0]
    theta = 34.767507632418315

    thicknesses = jnp.array(thickness_materials)
    n_k = utils_materials.get_n_k(materials, frequencies)

    thicknesses *= utils_units.get_nano()
    R_TE, T_TE, R_TM, T_TM = stackrt(n_k, thicknesses, frequencies, theta)

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    print("R_TE")
    for elem_1 in R_TE:
        for elem_2 in elem_1:
            print(elem_2)

    print("T_TE")
    for elem_1 in T_TE:
        for elem_2 in elem_1:
            print(elem_2)

    print("R_TM")
    for elem_1 in R_TM:
        for elem_2 in elem_1:
            print(elem_2)

    print("T_TM")
    for elem_1 in T_TM:
        for elem_2 in elem_1:
            print(elem_2)

    expected_R_avg = jnp.array([[0.48574877]])
    expected_T_avg = jnp.array([[0.32087105]])

    onp.testing.assert_allclose(R_avg, expected_R_avg)
    onp.testing.assert_allclose(T_avg, expected_T_avg)



import jax.numpy as jnp


def get_thicknesses_surrounded_by_air(thicknesses):
    assert isinstance(thicknesses, jnp.ndarray)
    assert thicknesses.ndim == 1

    return jnp.concatenate([jnp.array([0.0]), thicknesses, jnp.array([0.0])], axis=0)



import numpy as onp
import jax.numpy as jnp

from jaxlayerlumos import jaxlayerlumos


def compute_properties_jaxlayerlumos(thicknesses, n_k, frequencies, angle_of_incidence):
    assert isinstance(thicknesses, onp.ndarray)
    assert isinstance(n_k, onp.ndarray)
    assert isinstance(frequencies, onp.ndarray)
    assert isinstance(angle_of_incidence, float)

    assert thicknesses.ndim == 1
    assert n_k.ndim == 2
    assert frequencies.ndim == 1

    assert thicknesses.shape[0] == n_k.shape[0]
    assert frequencies.shape[0] == n_k.shape[1]

    n_k = jnp.array(n_k).T
    thicknesses = jnp.array(thicknesses)
    frequencies = jnp.array(frequencies)

    R_TE, T_TE, R_TM, T_TM = jaxlayerlumos.stackrt_n_k(
        n_k, thicknesses, frequencies, jnp.array([angle_of_incidence])
    )

    R_TE = onp.array(R_TE)
    T_TE = onp.array(T_TE)
    R_TM = onp.array(R_TM)
    T_TM = onp.array(T_TM)

    R_TE = onp.squeeze(R_TE, axis=0)
    T_TE = onp.squeeze(T_TE, axis=0)
    R_TM = onp.squeeze(R_TM, axis=0)
    T_TM = onp.squeeze(T_TM, axis=0)

    return R_TE, R_TM, T_TE, T_TM



import time
import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos.jaxlayerlumos import stackrt_n_k as stackrt_new
from jaxlayerlumos.jaxlayerlumos_old import stackrt as stackrt_old
from jaxlayerlumos import utils_materials
from jaxlayerlumos import utils_units


def compare_stackrt_old_new(use_zero_angle, use_thick_layers):
    wavelengths = jnp.linspace(300e-9, 900e-9, 10)
    frequencies = utils_units.get_light_speed() / wavelengths

    all_materials = utils_materials.get_all_materials()

    num_layerss = jnp.array([2, 4, 6, 8, 10])
    num_tests = 10

    random_state = onp.random.RandomState(42)

    times_new = []
    times_old = []

    counts_R_TE = 0
    counts_T_TE = 0
    counts_R_TM = 0
    counts_T_TM = 0
    counts_all = 0

    for num_layers in num_layerss:
        for _ in range(0, num_tests):
            materials = random_state.choice(all_materials, num_layers)

            if use_zero_angle:
                angle = 0.0
            else:
                angle = random_state.uniform(0.0, 89.9)
            materials = onp.concatenate([["Air"], materials], axis=0)

            n_k = utils_materials.get_n_k(materials, frequencies)

            if use_thick_layers:
                thickness_material = random_state.uniform(5.0, 100.0, num_layers)
            else:
                thickness_material = random_state.uniform(0.01, 10.0, num_layers)
            thicknesses = onp.insert(thickness_material, 0, 0)
            thicknesses[-1] = 0.0

            thicknesses *= utils_units.get_nano()

            thicknesses = jnp.array(thicknesses)

            time_start_old = time.monotonic()
            R_TE_old, T_TE_old, R_TM_old, T_TM_old = stackrt_old(
                n_k, thicknesses, frequencies, angle
            )
            time_end_old = time.monotonic()

            time_start_new = time.monotonic()
            R_TE_new, T_TE_new, R_TM_new, T_TM_new = stackrt_new(
                n_k, thicknesses, frequencies, angle
            )
            time_end_new = time.monotonic()

            try:
                R_TE_old = onp.clip(R_TE_old, min=1e-8)
                T_TE_old = onp.clip(T_TE_old, min=1e-8)
                R_TM_old = onp.clip(R_TM_old, min=1e-8)
                T_TM_old = onp.clip(T_TM_old, min=1e-8)

                R_TE_new = onp.clip(R_TE_new, min=1e-8)
                T_TE_new = onp.clip(T_TE_new, min=1e-8)
                R_TM_new = onp.clip(R_TM_new, min=1e-8)
                T_TM_new = onp.clip(T_TM_new, min=1e-8)
            except:
                R_TE_old = onp.clip(R_TE_old, a_min=1e-8, a_max=None)
                T_TE_old = onp.clip(T_TE_old, a_min=1e-8, a_max=None)
                R_TM_old = onp.clip(R_TM_old, a_min=1e-8, a_max=None)
                T_TM_old = onp.clip(T_TM_old, a_min=1e-8, a_max=None)

                R_TE_new = onp.clip(R_TE_new, a_min=1e-8, a_max=None)
                T_TE_new = onp.clip(T_TE_new, a_min=1e-8, a_max=None)
                R_TM_new = onp.clip(R_TM_new, a_min=1e-8, a_max=None)
                T_TM_new = onp.clip(T_TM_new, a_min=1e-8, a_max=None)

            str_thicknesses = [
                f"{thickness / utils_units.get_nano():.4f}"
                for thickness in thicknesses[1:]
            ]
            print(f":materials: [{', '.join(materials)}]")
            print(f":thicknesses: [{', '.join(str_thicknesses)}]")
            print(f":angle: {angle:.4f}")

            is_close_R_TE = onp.allclose(R_TE_old, R_TE_new, rtol=1e-5)
            is_close_T_TE = onp.allclose(T_TE_old, T_TE_new, rtol=1e-5)
            is_close_R_TM = onp.allclose(R_TM_old, R_TM_new, rtol=1e-5)
            is_close_T_TM = onp.allclose(T_TM_old, T_TM_new, rtol=1e-5)

            if not is_close_R_TE:
                print("R_TE does not match")
                print(R_TE_old)
                print(R_TE_new)
                counts_R_TE += 1
            if not is_close_T_TE:
                print("T_TE does not match")
                print(T_TE_old)
                print(T_TE_new)
                counts_T_TE += 1
            if not is_close_R_TM:
                print("R_TM does not match")
                print(R_TM_old)
                print(R_TM_new)
                counts_R_TM += 1
            if not is_close_T_TM:
                print("T_TM does not match")
                print(T_TM_old)
                print(T_TM_new)
                counts_T_TM += 1
            counts_all += 1

            time_consumed_new = time_end_new - time_start_new
            time_consumed_old = time_end_old - time_start_old

            print(f":new: {time_consumed_new:.4f} sec.")
            print(f":old: {time_consumed_old:.4f} sec.")
            if not (
                is_close_R_TE and is_close_T_TE and is_close_R_TM and is_close_T_TM
            ):
                print("=====NOT MATCHED=====")

            print("")

            times_new.append(time_consumed_new)
            times_old.append(time_consumed_old)

    if onp.mean(times_new) > onp.mean(times_new):
        assert False

    if counts_R_TE > 0 or counts_T_TE > 0 or counts_R_TM > 0 or counts_T_TM > 0:
        print("failure ratios")
        print(
            f"R_TE {counts_R_TE / counts_all:.4f} T_TE {counts_T_TE / counts_all:.4f}"
        )
        print(
            f"R_TM {counts_R_TM / counts_all:.4f} T_TM {counts_T_TM / counts_all:.4f}"
        )

        assert False


def test_comparison_stackrt_old_new_zero_angle_thin():
    use_zero_angle = True
    use_thick_layers = False

    compare_stackrt_old_new(use_zero_angle, use_thick_layers)


def test_comparison_stackrt_old_new_zero_angle_thick():
    use_zero_angle = True
    use_thick_layers = True

    compare_stackrt_old_new(use_zero_angle, use_thick_layers)


def test_comparison_stackrt_old_new_nonzero_angle_thin():
    use_zero_angle = False
    use_thick_layers = False

    compare_stackrt_old_new(use_zero_angle, use_thick_layers)


def test_comparison_stackrt_old_new_nonzero_angle_thick():
    use_zero_angle = False
    use_thick_layers = True

    compare_stackrt_old_new(use_zero_angle, use_thick_layers)



import numpy as np
import matplotlib.pyplot as plt
import os


def plot_spectra(
    frequencies, Rs_TE, Rs_TM, Ts_TE, Ts_TM, str_labels, linestyles, str_file
):
    assert frequencies.ndim == 1
    assert Rs_TE.ndim == 2
    assert Rs_TM.ndim == 2
    assert Ts_TE.ndim == 2
    assert Ts_TM.ndim == 2
    assert Rs_TE.shape[0] == Rs_TM.shape[0] == Ts_TE.shape[0] == Ts_TM.shape[0]
    assert Rs_TE.shape[0] == len(str_labels) == len(linestyles)
    assert Rs_TE.shape[1] == Rs_TM.shape[1] == Ts_TE.shape[1] == Ts_TM.shape[1]
    assert Rs_TE.shape[1] == frequencies.shape[0]

    fig, axs = plt.subplots(2, 2, figsize=(12, 12))

    for R_TE, R_TM, T_TE, T_TM, str_label, linestyle in zip(
        Rs_TE, Rs_TM, Ts_TE, Ts_TM, str_labels, linestyles
    ):
        axs[0, 0].plot(
            frequencies, R_TE, label=str_label, linestyle=linestyle, linewidth=4
        )
        axs[0, 1].plot(
            frequencies, R_TM, label=str_label, linestyle=linestyle, linewidth=4
        )
        axs[1, 0].plot(
            frequencies, T_TE, label=str_label, linestyle=linestyle, linewidth=4
        )
        axs[1, 1].plot(
            frequencies, T_TM, label=str_label, linestyle=linestyle, linewidth=4
        )

    fontsize = 18

    axs[0, 0].set_xlabel("Frequency (Hz)", fontsize=fontsize)
    axs[0, 0].set_ylabel("Reflectance_TE", fontsize=fontsize)
    axs[0, 0].legend(fontsize=fontsize)
    axs[0, 0].grid()

    axs[0, 1].set_xlabel("Frequency (Hz)", fontsize=fontsize)
    axs[0, 1].set_ylabel("Reflectance_TM", fontsize=fontsize)
    axs[0, 1].legend(fontsize=fontsize)
    axs[0, 1].grid()

    axs[1, 0].set_xlabel("Frequency (Hz)", fontsize=fontsize)
    axs[1, 0].set_ylabel("Transmittance_TE", fontsize=fontsize)
    axs[1, 0].legend(fontsize=fontsize)
    axs[1, 0].grid()

    axs[1, 1].set_xlabel("Frequency (Hz)", fontsize=fontsize)
    axs[1, 1].set_ylabel("Transmittance_TM", fontsize=fontsize)
    axs[1, 1].legend(fontsize=fontsize)
    axs[1, 1].grid()

    plt.tight_layout()
    plt.suptitle(str_file, fontsize=fontsize)
    plt.subplots_adjust(top=0.95)

    str_directory = "../assets/comparisons"
    if not os.path.exists(str_directory):
        os.mkdir(str_directory)

    plt.savefig(os.path.join(str_directory, str_file + ".png"))



import jax.numpy as jnp
import numpy as np

from jaxlayerlumos import utils_materials


def verify_values(frequencies, values):
    unique_frequencies, indices = jnp.unique(frequencies, return_index=True)
    unique_values = values[indices]

    print(frequencies[0], frequencies[-1])
    print(unique_frequencies[0], unique_frequencies[-1])
    print(values[0], values[-1])
    print(unique_values[0], unique_values[-1])

    np.testing.assert_allclose(unique_frequencies, frequencies)
    np.testing.assert_allclose(unique_values, values)

    indices = jnp.argsort(frequencies)
    sorted_frequencies = frequencies[indices]
    sorted_values = values[indices]

    print(frequencies[0], frequencies[-1])
    print(sorted_frequencies[0], sorted_frequencies[-1])
    print(values[0], values[-1])
    print(sorted_values[0], sorted_values[-1])

    np.testing.assert_allclose(sorted_frequencies, frequencies)
    np.testing.assert_allclose(sorted_values, values)


def test_material_values():
    all_materials = utils_materials.get_all_materials()

    for material in all_materials:
        n_material, k_material = utils_materials.load_material(material)

        print(f"{material} n")
        frequencies_n, values_n = n_material.T
        verify_values(frequencies_n, values_n)

        print(f"{material} k")
        frequencies_k, values_k = k_material.T
        verify_values(frequencies_k, values_k)



import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import utils_spectra


def test_convert_frequencies_to_wavelengths():
    num_frequencies = 101

    frequencies = jnp.linspace(1e10, 100000e10, num_frequencies)
    wavelengths = utils_spectra.convert_frequencies_to_wavelengths(frequencies)
    frequencies_ = utils_spectra.convert_wavelengths_to_frequencies(wavelengths)

    assert wavelengths.ndim == 1
    assert wavelengths.shape[0] == num_frequencies
    assert frequencies_.ndim == 1
    assert frequencies_.shape[0] == num_frequencies

    onp.testing.assert_allclose(frequencies_, frequencies)


def test_convert_wavelengths_to_frequencies():
    num_wavelengths = 101

    wavelengths = jnp.linspace(1e-9, 100000e-9, num_wavelengths)
    frequencies = utils_spectra.convert_wavelengths_to_frequencies(wavelengths)
    wavelengths_ = utils_spectra.convert_frequencies_to_wavelengths(frequencies)

    assert frequencies.ndim == 1
    assert frequencies.shape[0] == num_wavelengths
    assert wavelengths_.ndim == 1
    assert wavelengths_.shape[0] == num_wavelengths

    onp.testing.assert_allclose(wavelengths_, wavelengths)


def test_get_frequencies_visible_light():
    num_wavelengths = 1234

    frequencies = utils_spectra.get_frequencies_visible_light(
        num_wavelengths=num_wavelengths
    )

    assert frequencies.ndim == 1
    assert frequencies.shape[0] == num_wavelengths
    onp.testing.assert_allclose(onp.min(frequencies), 3.84349305e14)
    onp.testing.assert_allclose(onp.max(frequencies), 7.88927521e14)


def test_get_frequencies_wide_visible_light():
    num_wavelengths = 1234

    frequencies = utils_spectra.get_frequencies_wide_visible_light(
        num_wavelengths=num_wavelengths
    )

    assert frequencies.ndim == 1
    assert frequencies.shape[0] == num_wavelengths
    onp.testing.assert_allclose(onp.min(frequencies), 3.331027e14)
    onp.testing.assert_allclose(onp.max(frequencies), 9.993082e14)



import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import stackrt
from jaxlayerlumos import utils_radio_frequency


def test_stackrt_radio_frequency():
    frequencies = jnp.linspace(8e9, 18e9, 100)
    materials = ["Air", "Ag", "Air"]

    n_k_ag = utils_radio_frequency.load_material_RF("Ag", frequencies)
    n_ag = (
        n_k_ag[:, 1] + 1j * n_k_ag[:, 2]
    )  # Combine n and k into a complex refractive index

    n_air = jnp.ones_like(frequencies)  # Refractive index of air
    d_air = jnp.array([0])  # Air layer thickness
    d_ag = jnp.array([5e-8])  # Ag layer thickness

    n_stack = jnp.vstack([n_air, n_ag, n_air]).T  # Transpose to match expected shape
    d_stack = jnp.hstack([d_air, d_ag, d_air])  # Stack thickness

    R_TE, T_TE, R_TM, T_TM = stackrt(n_stack, d_stack, frequencies, jnp.array([0]))

    SE_TE = -10 * jnp.log10(T_TE)
    SE_TM = -10 * jnp.log10(T_TM)
    SE = (SE_TE + SE_TM) / 2

    expected_mean_SE = 55.47150237872992

    onp.testing.assert_allclose(jnp.mean(SE), expected_mean_SE)



import pytest
import jax.numpy as jnp
import numpy as onp
import scipy.constants as scic

from jaxlayerlumos import utils_units


def test_get_light_speed():
    onp.testing.assert_allclose(utils_units.get_light_speed(), scic.c)
    onp.testing.assert_allclose(utils_units.get_light_speed(), 299792458)


def test_get_nano():
    onp.testing.assert_allclose(utils_units.get_nano(), scic.nano)
    onp.testing.assert_allclose(utils_units.get_nano(), 1e-9)


def test_get_milli():
    onp.testing.assert_allclose(utils_units.get_milli(), scic.milli)
    onp.testing.assert_allclose(utils_units.get_milli(), 1e-3)


def test_get_centi():
    onp.testing.assert_allclose(utils_units.get_centi(), scic.centi)
    onp.testing.assert_allclose(utils_units.get_centi(), 1e-2)


def test_get_giga():
    onp.testing.assert_allclose(utils_units.get_giga(), scic.giga)
    onp.testing.assert_allclose(utils_units.get_giga(), 1e9)


def test_convert_nm_to_m():
    thicknesses = jnp.array([10.0, 20.0, 4.0, 1.0, 2.0])
    thicknesses_in_m = utils_units.convert_nm_to_m(thicknesses)

    assert thicknesses_in_m.ndim == 1
    assert thicknesses_in_m.shape[0] == thicknesses.shape[0]

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.nano)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-9)

    thicknesses = jnp.array(1000.0)
    thicknesses_in_m = utils_units.convert_nm_to_m(thicknesses)

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.nano)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-9)

    thickness = 1000.0
    thickness_in_m = utils_units.convert_nm_to_m(thickness)

    onp.testing.assert_allclose(thickness_in_m, thickness * scic.nano)
    onp.testing.assert_allclose(thickness_in_m, thickness * 1e-9)


def test_convert_mm_to_m():
    thicknesses = jnp.array([10.0, 20.0, 4.0, 1.0, 2.0])
    thicknesses_in_m = utils_units.convert_mm_to_m(thicknesses)

    assert thicknesses_in_m.ndim == 1
    assert thicknesses_in_m.shape[0] == thicknesses.shape[0]

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.milli)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-3)

    thicknesses = jnp.array(1000.0)
    thicknesses_in_m = utils_units.convert_mm_to_m(thicknesses)

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.milli)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-3)

    thickness = 1000.0
    thickness_in_m = utils_units.convert_mm_to_m(thickness)

    onp.testing.assert_allclose(thickness_in_m, thickness * scic.milli)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-3)


def test_convert_cm_to_m():
    thicknesses = jnp.array([10.0, 20.0, 4.0, 1.0, 2.0])
    thicknesses_in_m = utils_units.convert_cm_to_m(thicknesses)

    assert thicknesses_in_m.ndim == 1
    assert thicknesses_in_m.shape[0] == thicknesses.shape[0]

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.centi)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-2)

    thicknesses = jnp.array(1000.0)
    thicknesses_in_m = utils_units.convert_cm_to_m(thicknesses)

    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * scic.centi)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-2)

    thickness = 1000.0
    thickness_in_m = utils_units.convert_cm_to_m(thickness)

    onp.testing.assert_allclose(thickness_in_m, thickness * scic.centi)
    onp.testing.assert_allclose(thicknesses_in_m, thicknesses * 1e-2)



import numpy as onp
import jax.numpy as jnp
from jaxlayerlumos import stackrt

from jaxlayerlumos.utils_materials import get_n_k_surrounded_by_air
from jaxlayerlumos.utils_spectra import get_frequencies_visible_light
from jaxlayerlumos.utils_spectra import convert_frequencies_to_wavelengths
from jaxlayerlumos.utils_layers import get_thicknesses_surrounded_by_air
from jaxlayerlumos.utils_units import convert_nm_to_m


if __name__ == "__main__":
    frequencies = get_frequencies_visible_light()
    wavelengths = convert_frequencies_to_wavelengths(frequencies)

    list_materials = [
        ["Ag"],
        ["Ag"],
        ["Ag", "Al", "Ag"],
        ["TiO2", "Ag", "TiO2"],
    ]
    list_thicknesses = [
        jnp.array([100.0]),
        jnp.array([10.0]),
        jnp.array([10.0, 11.0, 12.0]),
        jnp.array([20.0, 5.0, 30.0]),
    ]
    angles = jnp.array([0.0, 45.0, 75.0, 89.0])

    for angle in angles:
        for materials, thicknesses in zip(list_materials, list_thicknesses):
            assert len(materials) == thicknesses.shape[0]

            n_k = get_n_k_surrounded_by_air(materials, frequencies)
            layers = convert_nm_to_m(get_thicknesses_surrounded_by_air(thicknesses))

            R_TE, T_TE, R_TM, T_TM = stackrt(
                n_k, layers, frequencies, jnp.array([angle])
            )

            R_TE_jll = onp.squeeze(onp.array(R_TE), axis=0)
            R_TM_jll = onp.squeeze(onp.array(R_TM), axis=0)
            T_TE_jll = onp.squeeze(onp.array(T_TE), axis=0)
            T_TM_jll = onp.squeeze(onp.array(T_TM), axis=0)

            print(R_TE_jll.shape)
            print(R_TM_jll.shape)
            print(T_TE_jll.shape)
            print(T_TM_jll.shape)

            spectrum_reflection = (R_TE_jll + R_TM_jll) / 2
            spectrum_tranmission = (T_TE_jll + T_TM_jll) / 2

            print(spectrum_reflection.shape)
            print(spectrum_tranmission.shape)



import jax.numpy as jnp

from jaxlayerlumos import utils_units


def convert_frequencies_to_wavelengths(frequencies):
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1

    wavelengths = utils_units.get_light_speed() / frequencies
    return wavelengths


def convert_wavelengths_to_frequencies(wavelengths):
    assert isinstance(wavelengths, jnp.ndarray)
    assert wavelengths.ndim == 1

    frequencies = utils_units.get_light_speed() / wavelengths
    return frequencies


def get_frequencies_visible_light(num_wavelengths=1001):
    wavelengths = jnp.linspace(
        380 * utils_units.get_nano(), 780 * utils_units.get_nano(), num_wavelengths
    )
    frequencies = convert_wavelengths_to_frequencies(wavelengths)

    return frequencies


def get_frequencies_wide_visible_light(num_wavelengths=1001):
    wavelengths = jnp.linspace(
        300 * utils_units.get_nano(), 900 * utils_units.get_nano(), num_wavelengths
    )
    frequencies = convert_wavelengths_to_frequencies(wavelengths)

    return frequencies



import jax.numpy as jnp
import jax
import matplotlib.pyplot as plt

from jaxlayerlumos import stackrt
from jaxlayerlumos.utils_materials import (
    get_n_k_surrounded_by_air,
    get_n_k_surrounded_by_air_and_pec,
)
from jaxlayerlumos.utils_layers import (
    get_thicknesses_surrounded_by_air,
    convert_mm_to_m,
)

if __name__ == "__main__":
    # Define frequency range (0.1 GHz to 1 GHz)
    frequencies = jnp.arange(0.1e9, 1.01e9, 0.01e9)  # Frequencies from 0.1 GHz to 1 GHz
    # frequencies = jnp.array([1e9])
    # materials = ["Mat1", "Mat2", "Mat6", "Mat7", "Mat8"]  # Material list
    # materials = ["Mat16", "Mat6", "Mat5", "Mat11", "Mat2"]  # Material list
    # materials = ["Mat11", "Mat16", "Mat7", "Mat4", "Mat4"] # Material list
    materials = ["Mat6", "Mat1", "Mat6"]  # Material list
    # thicknesses = jnp.array([1.155, 0.885, 1.272, 1.446, 0.486])  # Thickness in mm
    thicknesses = jnp.array([2.0, 1.0, 1.0])  # Thickness in mm
    angles = 45.0  # Incident angles in degrees
    # angles = jnp.array([0.0, 45.0, 75.0, 89.0])  # Incident angles in degrees

    # Retrieve refractive indices for all frequencies at once
    # n_k_all = get_n_k_surrounded_by_air(materials, frequencies)
    n_k_all = get_n_k_surrounded_by_air_and_pec(materials, frequencies)

    # Inspect refractive indices
    print("Refractive indices (n_k_all):", n_k_all)

    # Convert thicknesses to meters
    layers = convert_mm_to_m(get_thicknesses_surrounded_by_air(thicknesses))

    R_TE, T_TE, R_TM, T_TM = stackrt(n_k_all, layers, frequencies, angles)
    # Inspect reflectance values before dB conversion
    print("Reflectance values (R_TE_values):", R_TE)

    # Ensure R_TE_values are valid for dB conversion
    R_TE_db = jnp.squeeze(10 * jnp.log10(R_TE))
    R_TM_db = jnp.squeeze(10 * jnp.log10(R_TM))

    # Plot reflection over frequency in dB
    plt.figure(figsize=(10, 6))
    plt.semilogx(
        frequencies * 1e-9, R_TE_db, label="Reflectance (TE) in dB"
    )  # Convert frequencies to GHz for x-axis
    plt.semilogx(
        frequencies * 1e-9, R_TM_db, label="Reflectance (TM) in dB"
    )  # Convert frequencies to GHz for x-axis
    plt.xlabel("Frequency (GHz)")
    plt.ylabel("Reflectance (dB)")
    plt.title("Reflectance vs Frequency")
    plt.legend()
    plt.grid(True)
    plt.show()



import jax.numpy as jnp
import numpy as np


def calc_position_in_structure(thickness_materials, position):

    cs = np.cumsum(thickness_materials[:-1])
    layer = np.sum(position[:, None] >= cs, axis=1)
    # layer = np.sum(position >= cs, axis=0)
    distance = (position - cs[layer - 1]).T
    return [layer, distance]


def calc_position_data(layer, position_in_layer, results, update_results=True):
    coeff_TE = results["coeff_TE"][:, :, layer, :]
    coeff_TM = results["coeff_TM"][:, :, layer, :]
    n = results["n"][:, :, layer]
    kz = results["k_z"][:, :, layer]
    cos_theta = results["cos_theta"][:, :, layer]

    n0 = results["n"][:, :, 0]
    cos_theta0 = results["cos_theta"][:, :, 0]

    n0 = n0[:, :, np.newaxis]
    cos_theta0 = cos_theta0[:, :, np.newaxis]

    Ef_TE = coeff_TE[:, :, :, 0] * np.exp(1j * kz * position_in_layer)
    Eb_TE = coeff_TE[:, :, :, 1] * np.exp(-1j * kz * position_in_layer)

    poyn_TE = (
        np.real(n * cos_theta * np.conj(Ef_TE + Eb_TE) * (Ef_TE - Eb_TE))
    ) / np.real(n0 * cos_theta0)

    Ef_TM = coeff_TM[:, :, :, 0] * np.exp(1j * kz * position_in_layer)
    Eb_TM = coeff_TM[:, :, :, 1] * np.exp(-1j * kz * position_in_layer)

    poyn_TM = (
        np.real(n * np.conj(cos_theta) * (Ef_TM + Eb_TM) * np.conj(Ef_TM - Eb_TM))
    ) / (n0 * np.conj(cos_theta0))

    absorb_TE = np.imag(n * cos_theta * kz * np.abs(Ef_TE + Eb_TE) ** 2) / np.real(
        n0 * cos_theta0
    )
    absorb_TM = np.imag(
        n
        * np.conj(cos_theta)
        * (kz * np.abs(Ef_TM - Eb_TM) ** 2 - np.conj(kz) * np.abs(Ef_TM + Eb_TM) ** 2)
    ) / np.real(n0 * np.conj(cos_theta0))

    E_TE = [0, Ef_TE + Eb_TE, 0]
    E_TM = [
        (Ef_TM - Eb_TM) * cos_theta,
        0,
        (-Ef_TM - Eb_TM) * np.sqrt(1 - cos_theta**2),
    ]

    results_position = {
        "poyn_TE": poyn_TE,
        "poyn_TM": poyn_TM,
        "absorb_TE": absorb_TE,
        "absorb_TM": absorb_TM,
        "E_TE": E_TE,
        "E_TM": E_TM,
    }
    if update_results:
        results.update(results_position)
        return results
    else:
        return results_position


def calc_absorption_in_each_layer(thicknesses, results):
    # Assuming d, power_entering_TE, power_entering_TM, t_TE_i, t_TM_i, kz_layers,
    # cos_theta, n_layers, coeff_TE, coeff_TM, and calc_position_data are already defined.

    power_entering_TE = results["power_entering_TE"]
    power_entering_TM = results["power_entering_TM"]
    num_layers = len(thicknesses)

    power_entering_each_layer_TE = np.zeros((num_layers, power_entering_TE.size))
    power_entering_each_layer_TM = np.zeros((num_layers, power_entering_TM.size))

    # Set initial conditions
    power_entering_each_layer_TE[0, :] = 1
    power_entering_each_layer_TM[0, :] = 1

    power_entering_each_layer_TE[1, :] = power_entering_TE
    power_entering_each_layer_TM[1, :] = power_entering_TM

    power_entering_each_layer_TE[-1, :] = results["T_TE"]
    power_entering_each_layer_TM[-1, :] = results["T_TM"]

    # Loop for intermediate layers
    for i in range(2, num_layers - 1):
        results_out = calc_position_data([i], 0, results, update_results=False)

        power_entering_each_layer_TE[i, :] = np.real(results_out["poyn_TE"][0, :, 0])
        power_entering_each_layer_TM[i, :] = np.real(results_out["poyn_TM"][0, :, 0])

    # Calculate absorption
    absorption_TE = -np.diff(power_entering_each_layer_TE, axis=0)
    absorption_TM = -np.diff(power_entering_each_layer_TM, axis=0)

    # Add the final row to absorption (equivalent to MATLAB's behavior)
    absorption_TE = np.vstack((absorption_TE, power_entering_each_layer_TE[-1, :]))
    absorption_TM = np.vstack((absorption_TM, power_entering_each_layer_TM[-1, :]))

    results.update(
        {"absorption_layer_TE": absorption_TE, "absorption_layer_TM": absorption_TM}
    )
    return results



import jax.numpy as jnp

from jaxlayerlumos.jaxlayerlumos import stackrt_n_k


def stackrt_n_k_0(n, d, f):
    return stackrt_n_k(n, d, f, jnp.array([0]))


def stackrt_n_k_45(n, d, f):
    return stackrt_n_k(n, d, f, jnp.array([45]))



import jax.numpy as jnp
import numpy as onp
import scipy.constants as scic

from jaxlayerlumos import stackrt
from jaxlayerlumos import utils_materials
from jaxlayerlumos import utils_units


def test_stackrt_1():
    wavelengths = jnp.linspace(300e-9, 900e-9, 3)
    frequencies = scic.c / wavelengths
    # frequencies = 4e9

    materials = onp.array(["Air", "Ag", "Air"])
    thickness_materials = [0, 2e-6, 0]

    thicknesses = jnp.array(thickness_materials)
    n_k = utils_materials.get_n_k(materials, frequencies)

    R_TE, T_TE, R_TM, T_TM = stackrt(n_k, thicknesses, frequencies, 0.0)

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    expected_R_avg = onp.array(
        [
            [
                0.09525774381746192,
                0.960112317038949,
                0.9772199561956647,
            ],
        ]
    )
    expected_T_avg = onp.array(
        [
            [
                7.770517514983971e-25,
                1.7609707728810983e-65,
                1.4500794728372322e-70,
            ],
        ]
    )

    print("R_avg")
    for elem in R_avg:
        for elem2 in elem:
            print(elem2)
    print("T_avg")
    for elem in T_avg:
        for elem2 in elem:
            print(elem2)

    onp.testing.assert_allclose(R_avg, expected_R_avg)
    onp.testing.assert_allclose(T_avg, expected_T_avg)


def test_stackrt_2():
    wavelengths = jnp.linspace(300e-9, 900e-9, 3)
    frequencies = utils_units.get_light_speed() / wavelengths

    materials = onp.array(["Air", "FusedSilica", "Si3N4"])
    thickness_materials = [0, 2.91937911, 0]

    n_k = utils_materials.get_n_k(materials, frequencies)

    thicknesses = jnp.array(thickness_materials)

    R_TE, T_TE, R_TM, T_TM = stackrt(n_k, thicknesses, frequencies, 0.0)

    R_avg = (R_TE + R_TM) / 2
    T_avg = (T_TE + T_TM) / 2

    expected_R_avg = onp.array(
        [
            [
                0.00013699548845149838,
                0.09585152519397383,
                0.0671778954615817,
            ],
        ]
    )
    expected_T_avg = onp.array(
        [
            [
                0.9998630045115486,
                0.904148474806026,
                0.9328221045384182,
            ],
        ]
    )

    print("R_avg")
    for elem in R_avg:
        for elem2 in elem:
            print(elem2)
    print("T_avg")
    for elem in T_avg:
        for elem2 in elem:
            print(elem2)

    onp.testing.assert_allclose(R_avg, expected_R_avg)
    onp.testing.assert_allclose(T_avg, expected_T_avg)



import jax.numpy as jnp
import numpy as onp
import os
import csv
import json
from pathlib import Path

from jaxlayerlumos import utils_spectra
from jaxlayerlumos import utils_radar_materials


def load_json():
    current_dir = str(Path(__file__).parent)
    materials_file = os.path.join(current_dir, "materials.json")

    with open(materials_file, "r") as file_json:
        material_indices = json.load(file_json)

    return material_indices, current_dir


def get_all_materials():
    material_indices, _ = load_json()
    return list(material_indices.keys())


def load_material_wavelength_um(material):
    material_indices, str_directory = load_json()
    str_file = material_indices.get(material)

    if not str_file:
        raise ValueError(f"Material {material} not found in JaxLayerLumos.")

    str_csv = os.path.join(str_directory, str_file)
    data_n = []
    data_k = []

    with open(str_csv, "r") as csvfile:
        csvreader = csv.reader(csvfile)

        start_n = False
        start_k = False

        for row in csvreader:
            if len(row) == 2:
                if row[0] == "wl" and row[1] == "n":
                    start_n = True
                    start_k = False
                elif row[0] == "wl" and row[1] == "k":
                    start_n = False
                    start_k = True
                else:
                    wavelength_um, value = map(float, row)

                    if start_n and not start_k:
                        data_n.append([wavelength_um, value])
                    elif not start_n and start_k:
                        data_k.append([wavelength_um, value])
                    else:
                        raise ValueError
            elif len(row) == 0:
                pass
            else:
                raise ValueError

    data_n = jnp.array(data_n)
    data_k = jnp.array(data_k)
    assert data_n.shape[0] > 0 or data_k.shape[0] > 0

    if data_n.shape[0] == 0:
        data_n = jnp.concatenate(
            [data_k[:, 0][..., jnp.newaxis], jnp.zeros((data_k.shape[0], 1))], axis=1
        )
    if data_k.shape[0] == 0:
        data_k = jnp.concatenate(
            [data_n[:, 0][..., jnp.newaxis], jnp.zeros((data_n.shape[0], 1))], axis=1
        )

    return data_n, data_k


def load_material_wavelength(material):
    data_n, data_k = load_material_wavelength_um(material)

    data_n = data_n.at[:, 0].set(data_n[:, 0] * 1e-6)
    data_k = data_k.at[:, 0].set(data_k[:, 0] * 1e-6)

    return data_n, data_k


def load_material(material):
    data_n, data_k = load_material_wavelength(material)

    data_n = data_n.at[:, 0].set(
        utils_spectra.convert_wavelengths_to_frequencies(data_n[:, 0])
    )
    data_k = data_k.at[:, 0].set(
        utils_spectra.convert_wavelengths_to_frequencies(data_k[:, 0])
    )

    return data_n, data_k


def interpolate(freqs_values, frequencies):
    assert isinstance(freqs_values, jnp.ndarray)
    assert isinstance(frequencies, jnp.ndarray)
    assert freqs_values.ndim == 2
    assert frequencies.ndim == 1

    freqs, values = freqs_values.T

    assert jnp.min(freqs) * 0.70 <= jnp.min(frequencies)
    assert jnp.max(frequencies) <= jnp.max(freqs) * 1.30

    values_interpolated = jnp.interp(
        frequencies,
        freqs,
        values,
        left="extrapolate",
        right="extrapolate",
    )

    return values_interpolated


def interpolate_material_n_k(material, frequencies):
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1

    if material == "Air":
        n_material = jnp.ones_like(frequencies)
        k_material = jnp.zeros_like(frequencies)
    else:
        data_n, data_k = load_material(material)
        n_material = interpolate(data_n, frequencies)
        k_material = interpolate(data_k, frequencies)

    return n_material, k_material


def get_eps_mu(materials, frequencies):
    assert isinstance(materials, (list, onp.ndarray))
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1
    assert materials[0] == "Air"

    materials = onp.array(materials)
    eps_r, mu_r = utils_radar_materials.get_eps_mu_Michielssen(
        materials[1:-1].astype(int), frequencies
    )

    n_k_air = get_n_k(materials[:1], frequencies)
    n_k_air = n_k_air.T
    eps_air, mu_air = convert_n_k_to_eps_mu_for_non_magnetic_materials(n_k_air)

    if materials[-1] == "PEC":
        eps_last = jnp.zeros_like(eps_air) + jnp.inf
        mu_last = jnp.ones_like(eps_air)
    else:
        try:
            eps_last, mu_last = utils_radar_materials.get_eps_mu_Michielssen(
                materials[-1:].astype(int), frequencies
            )
        except:
            raise NotImplementedError("This condition is not implemented yet.")

    eps_r = jnp.concatenate([eps_air, eps_r, eps_last], axis=0)
    mu_r = jnp.concatenate([mu_air, mu_r, mu_last], axis=0)

    eps_r = eps_r.T
    mu_r = mu_r.T

    return eps_r, mu_r


def interpolate_material_eps_mu(material, frequencies):
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1

    data_eps_r, data_mu_r = load_material_f_ghz(material)

    eps_r_real = interpolate(jnp.real(data_eps_r), frequencies)
    eps_r_imag = interpolate(jnp.imag(data_eps_r), frequencies)

    mu_r_real = interpolate(jnp.real(data_eps_r), frequencies)
    mu_r_imag = interpolate(jnp.imag(data_eps_r), frequencies)

    return eps_r_real, eps_r_imag, mu_r_real, mu_r_imag


def get_n_k(materials, frequencies):
    assert isinstance(materials, (list, onp.ndarray))
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1

    num_layers = len(materials)
    num_frequencies = frequencies.shape[0]

    n_k = jnp.ones((num_layers, num_frequencies), dtype=jnp.complex128)

    for ind, material in enumerate(materials):
        n_material, k_material = interpolate_material_n_k(material, frequencies)
        n_k = n_k.at[ind, :].set(n_material + 1j * k_material)

    n_k = n_k.T
    return n_k


def get_n_k_surrounded_by_air(materials, frequencies):
    assert isinstance(materials, (list, onp.ndarray))
    assert isinstance(frequencies, jnp.ndarray)
    assert frequencies.ndim == 1

    n_k = get_n_k(onp.concatenate([["Air"], materials, ["Air"]], axis=0), frequencies)
    return n_k


def convert_n_k_to_eps_mu_for_non_magnetic_materials(n_k):
    eps = jnp.conj(n_k**2)
    mu = jnp.ones_like(eps)

    return eps, mu



import scipy.constants as scic


def get_light_speed():
    return scic.c


def get_nano():
    return scic.nano


def get_milli():
    return scic.milli


def get_centi():
    return scic.centi


def get_giga():
    return scic.giga


def convert_nm_to_m(thicknesses):
    return thicknesses * get_nano()


def convert_mm_to_m(thicknesses):
    return thicknesses * get_milli()


def convert_cm_to_m(thicknesses):
    return thicknesses * get_centi()



import jax.numpy as jnp
import numpy as onp

from jaxlayerlumos import utils_units


def get_eps_mu_Michielssen(material_indices, frequencies):
    assert isinstance(material_indices, onp.ndarray)
    assert isinstance(frequencies, jnp.ndarray)
    assert material_indices.ndim == 1
    assert frequencies.ndim == 1
    for material_index in material_indices:
        assert material_index in onp.arange(1, 17)

    # Gets parameters from Michiellsen
    f = frequencies / utils_units.get_giga()  # in GHz
    M_epsr = jnp.vstack(
        [
            jnp.tile(
                jnp.array([10, 50, 15, 15, 15])[:, None], (1, len(f))
            ),  # Materials 1 to 5
            jnp.array(
                [  # Frequency-dependent permittivity for materials 6 to 8
                    5 / (f**0.861) - 1j * (8 / (f**0.569)),
                    8 / (f**0.778) - 1j * (10 / (f**0.682)),
                    10 / (f**0.778) - 1j * (6 / (f**0.861)),
                ]
            ),
            jnp.full((8, len(f)), 15, dtype=complex),  # Materials 9 to 16
        ]
    )

    # Fill constant values for permeability (mur)
    M_mur = jnp.vstack(
        [
            jnp.ones((2, len(f))),  # Materials 1 and 2
            jnp.array(
                [  # Frequency-dependent permeability for materials 3 to 5
                    5 / (f**0.974) - 1j * (10 / (f**0.961)),
                    3 / (f**1.0) - 1j * (15 / (f**0.957)),
                    7 / (f**1.0) - 1j * (12 / (f**1.0)),
                ]
            ),
            jnp.ones((3, len(f))),  # Materials 6 to 8
            jnp.array(
                [  # Frequency-dependent permeability for materials 9 to 16
                    (35 * (0.8**2)) / (f**2 + 0.8**2)
                    - 1j * (35 * 0.8 * f) / (f**2 + 0.8**2),
                    (35 * (0.5**2)) / (f**2 + 0.5**2)
                    - 1j * (35 * 0.5 * f) / (f**2 + 0.5**2),
                    (30 * (1**2)) / (f**2 + 1**2) - 1j * (30 * f) / (f**2 + 1**2),
                    (18 * (0.5**2)) / (f**2 + 0.5**2)
                    - 1j * (18 * 0.5 * f) / (f**2 + 0.5**2),
                    (20 * (1.5**2)) / (f**2 + 1.5**2)
                    - 1j * (20 * 1.5 * f) / (f**2 + 1.5**2),
                    (30 * (2.5**2)) / (f**2 + 2.5**2)
                    - 1j * (30 * 2.5 * f) / (f**2 + 2.5**2),
                    (30 * (2**2)) / (f**2 + 2**2) - 1j * (30 * 2 * f) / (f**2 + 2**2),
                    (25 * (3.5**2)) / (f**2 + 3.5**2)
                    - 1j * (25 * 3.5 * f) / (f**2 + 3.5**2),
                ]
            ),
        ]
    )

    # Initialize epsr and mur for the given material_indices
    eps_r = M_epsr[material_indices - 1, :]  # Python uses 0-based indexing
    mu_r = M_mur[material_indices - 1, :]

    return eps_r, mu_r



import jax.numpy as jnp
import jax

from jaxlayerlumos import stackrt
from jaxlayerlumos.utils_materials import get_n_k_surrounded_by_air
from jaxlayerlumos.utils_spectra import get_frequencies_visible_light
from jaxlayerlumos.utils_layers import get_thicknesses_surrounded_by_air
from jaxlayerlumos.utils_units import convert_nm_to_m


if __name__ == "__main__":
    frequencies = get_frequencies_visible_light()
    materials = ["TiO2", "Ag", "TiO2"]
    thicknesses = jnp.array([20.0, 5.0, 30.0])
    angles = jnp.array([0.0, 45.0, 75.0, 89.0])

    n_k = get_n_k_surrounded_by_air(materials, frequencies)

    def R_TE(inputs):
        layers = convert_nm_to_m(get_thicknesses_surrounded_by_air(inputs))
        R_TE, _, _, _ = stackrt(n_k, layers, frequencies, angles)
        return R_TE[0, 0]

    gradients_R_TE = jax.grad(R_TE)(thicknesses)
    print(gradients_R_TE)



def test_import_jaxlayerlumos():
    import jaxlayerlumos


def test_import_stackrt():
    from jaxlayerlumos import stackrt
    from jaxlayerlumos import stackrt0
    from jaxlayerlumos import stackrt45

    assert callable(stackrt)
    assert callable(stackrt0)
    assert callable(stackrt45)


def test_import_jaxlayerlumos_stackrt():
    from jaxlayerlumos.jaxlayerlumos import stackrt_eps_mu
    from jaxlayerlumos.jaxlayerlumos import stackrt_eps_mu_theta

    from jaxlayerlumos.jaxlayerlumos import stackrt_n_k

    assert callable(stackrt_eps_mu)
    assert callable(stackrt_eps_mu_theta)

    assert callable(stackrt_n_k)


def test_import_utils():
    import jaxlayerlumos.utils_spectra
    import jaxlayerlumos.utils_materials
    import jaxlayerlumos.utils_radar_materials
    import jaxlayerlumos.utils_layers
    import jaxlayerlumos.utils_units


def test_import_wrappers():
    import jaxlayerlumos.wrappers
